<copyright> TESLA mu.TomFile extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: TomFile.t,v 1.17 1998/07/27 15:25:50 tiggr Exp $</id>
    </copyright>

implementation class
mu.TomFile extension tesla: tesla

end;

implementation instance
mu.TomFile extension tesla
{
  <doc> The top-level <c>constructs</c>.  </doc>
  MutableObjectArray foreign_code;

  <doc> The name of the output file.  Only set if the output needs to be
      written.  </doc>
  String output_filename;
}

void
  addForeignCode ForeignCode c
{
  if (!foreign_code)
    foreign_code = [MutableObjectArray with c];
  else
    [foreign_code add c];
}

void
  loadFile
{
  if (self == builtin_file)
    {
      [self need_semantics];
      return;
    }

  LexTom l = [[LexTom alloc] initWithStream [File open filename input: YES
						  flags: FILE_NOT_EXIST_RAISE]
			     named filename];
  ParseTom p = [ParseTom parser self using l];
  if (verbose_parse_file)
    [[[Parser report] print ("reading ", filename)] nl];
  [p parse];
  [l close];
}

void
  makeExtensionPerform tom.Invocation inv
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [inv fireAt elements[i]];
}

void
  makeMethodsPerform tom.Invocation inv
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [elements[i] makeMethodsPerform inv];
}

<doc> Return {TRUE} iff full semantics of this file need to be read.  This
    is only asked once, just before the parser digests its first token.
    </doc>
boolean
  need_semantics
{
  if (!single_unit || unit == single_unit)
    {
      MutableByteString fn = [MutableByteString new];
      String bn = [File basename filename without-extension: ".t"];

      if (!single_unit)
	[fn print ([unit name], "-", bn, ".c")];
      else
	[fn print (bn, ".c")];

      // If we would cache information about this file (esp. what happens
      // in the method bodies), we could skip reading it, since we could
      // use the cached information for the RC and RH file.
      // Fri Jul 17 22:36:59 1998, tiggr@gerbil.org
      if (flag_force_output
	  || [File file-needs-updating fn with-respect-to filename])
	output_filename = fn;

      = TRUE;
    }
}

<doc> Pass this to every of our {elements}.  </doc>
void
  singleUnitLoaded boolean second_pass
{
  if (verbose_single_file)
    [[[ParseTom report] print (second_pass ? "2 " : "1 ", filename)] nl];

  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [elements[i] singleUnitLoaded second_pass];
}

<doc> Write the output file, if needed.  If not needed, we act as if we
    write the output file, since the act of emitting the output file
    collects a lot of information that is needed for the RC and RH files.
    </doc>
void
  writeCFile
{
  TeslaOutput of;

  if (output_filename != nil)
    of = [[TeslaOutput alloc] init [File open output_filename output: YES
					 flags: (FILE_NOT_EXIST_CREATE
						 | FILE_EXIST_TRUNCATE)]
			      bufferSize 16384];
  else
    {
      // Maybe TeslaOutput should be an extension of all OutputStreams,
      // including Sink...
      // Fri Jul 17 22:57:25 1998, tiggr@gerbil.org
      of = [[TeslaOutput alloc] init [Sink new]];
    }

  if (!flag_readable_c)
    [of print ("# 1 \"", filename, "\"\n")];
  [[of print "/* Your mother was a gorilla.  */\n"]
    print ("#define TESLA_OUTPUT\n#include <", [unit name], "-r.h>\n")];

  unwind ([[[of print '\n'] flushOutput] close])
    ({
      int i, n;

      if (foreign_code != nil)
	for ({i = 0; n = [foreign_code length];}; i < n; i++)
	  [[foreign_code[i] precompile] writeCFile of];

      n = [elements length];
      for (i = 0; i < n; i++)
	[Extension (elements[i]) writeCFile of];
    });
}

<doc> Pass to the {elements}.  </doc>
void
  writeRCFile OutputStream of
	 pass int pass
      collect MutableArray collect
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [elements[i] writeRCFile of pass pass collect collect];
}

<doc> Pass to the {elements}.  </doc>
void
  writeRHFile OutputStream of
	 pass int pass
{
  int i, n = [elements length];
  for (i = 0; i < n; i++)
    [elements[i] writeRHFile of pass pass];
}

end; 
