<copyright> TESLA non-mu extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: extensions.t,v 1.4 1998/07/24 15:01:53 tiggr Exp $</id>
    </copyright>

<c>
#include <string.h>
#include <sys/stat.h>
</c>

implementation class
TeslaOutput: BufferedStream, tesla
{
  <doc> Nine strings used to indent.  </doc>
  static Array indents;

  <doc> Some strings used to new-line.  </doc>
  static Array newlines;

  <doc> The number of strings in the {newlines}.  </doc>
  const NUM_NEWLINES = 10;
}

void
  load Array arguments
{
  indents = [ObjectArray with ("", " ", "  ", "   ", "    ", "     ",
			       "      ", "       ", "\t")];

  MutableByteString s = [MutableByteString withCapacity NUM_NEWLINES];
  MutableArray nl = [MutableObjectArray withCapacity NUM_NEWLINES];
  [nl add nil];
  int i;
  for (i = 1; i < NUM_NEWLINES; i++)
    {
      [s add '\n'];
      [nl add [s copy]];
    }
  newlines = nl;
}

end;

implementation instance
TeslaOutput
{
  <doc> Current indentation.  </doc>
  int indent;

  <doc> Our current idea of the line number.  </doc>
  int line_number;

  <doc> And the corresponding file.  </doc>
  String filename;
}
  
<doc> Adjust {indent}.  </doc>
void
  indent int delta
{
  indent += delta;
}

id (self)
  indent
{
  int i = indent;
  while (i > 8)
    {
      [indents[8] write self];
      i -= 8;
    }
  [indents[i] write self];
}

<doc> Do not flush.  </doc>
id
  nl
{
  if (!flag_readable_c)
    return [self print ' '];

  [self print '\n'];
  = [self indent];
}

id
  nl Common e
{
  if (flag_readable_c)
    {
      [self print '\n'];
      return [self indent];
    }

  int l = !e ? 0 : [e line];
  String fn = !e ? nil : [e containing_file];
  if (!l || !fn || (l == line_number && fn == filename))
    return [self print ' '];

  if (fn == filename && line_number > 0
      && l > line_number && l < line_number + NUM_NEWLINES)
    [self print String (newlines[l - line_number])];
  else
    {
      [self print ("\n# ", l, " \"", fn, "\"\n")];
      filename = fn;
    }

  line_number = l;
  = [self indent];
}

end;

/******************** File (tesla) ********************/

implementation class
File extension tesla

<doc> Return {TRUE} if the {output} is outdated with respect to the
    {input}, i.e., if {output} does not exist or is older than {input}.
    </doc>
boolean (result)
  file-needs-updating String output
      with-respect-to String input
{
  pointer op, ip;
  int olen, ilen;

  (op, olen) = [ByteString (output) byteStringContents];
  (ip, ilen) = [ByteString (input) byteStringContents];

  <c> {
    struct stat sti, sto;
    char *os = memcpy (alloca (1 + olen), op, olen);

    os[olen] = 0;
    if (stat (os, &sto) < 0)
      result = 1;
    else
      {
	char *is = memcpy (alloca (1 + ilen), ip, ilen);
	is[ilen] = 0;

	/* No update needed if the input does not exist.  */
	if (!stat (is, &sti) && sto.st_mtime < sti.st_mtime)
	  result = 1;
      }
  } </c>
}

end;

implementation instance
File extension tesla

end;

/******************** LexTom (tesla) ********************/

implementation class
LexTom extension tesla

end;

implementation instance
LexTom extension tesla
{
  <doc> Our parser is a different kind.  </doc>
  redeclare ParseTom parser;
}

<doc> Return {TRUE} for /c/ tags while the parser needs the semantics.
    </doc>
boolean
  docp String tag
{
  = [tag length] == 1 && tag[0] == 'c' && [parser need_semantics];
}

<doc> Pass the body to the {parser}.  </doc>
int
  documentation String body
	    tag String tag
	     at int line
pre
  [tag equal "c"]
{
  = [parser cToken body at line];
}

end;
